﻿using System;
using System.Web;
using System.Data;
using System.Data.OleDb;
using System.IO;

namespace JWorldBuilder.HTMLTool
{
	/// <summary>
	/// HTML lapok generálására használható osztály.
	/// </summary>
  public class HTML
  {
    /// <summary>
    /// HTML lapok generálására használható osztály.
    /// </summary>
    public HTML()
    {
    }
    
    //----------------------------------------------------------------------

    #region IfNotNull
    /// <summary>
    /// Jól használható függvény. Ha a Text nem null a NotNullText-et adja vissza.
    /// Ha a Text null akkor ezt adja vissza: "".
    /// </summary>
    /// <param name="Text">Az ellenörzendo szöveg.</param>
    /// <param name="NotNullText">Ha nem null ezt adja vissza a függvény.</param>
    /// <returns></returns>
    public static string IFNOTNULL(string Text,
                                   string NotNullText)
    {
      if(Text != null)
      {
        return NotNullText;
      }
      return "";
    }
    #endregion

    //----------------------------------------------------------------------

    #region Document Tags

    #region HTML Tag
    /// <summary>
    /// Ezzel kell kezdeni a HTML file-t.
    /// </summary>
    /// <returns></returns>
    public static string HtmlStart()
    {
      return "<HTML>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ezzel kell befejezni a HTML File-t.
    /// </summary>
    /// <returns></returns>
    public static string HtmlEnd()
    {
      return "</HTML>\n";
    }
    #endregion

    //----------------------------------------------------------------------

    #region HEAD Tag
    /// <summary>
    /// Ezzel kell kezdeni a dokumentum header-jét.
    /// </summary>
    /// <returns></returns>
    public static string HeadStart()
    {
      return "<HEAD>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ezzel kell befejezni a dokumentum header-jét.
    /// </summary>
    /// <returns></returns>
    public static string HeadEnd()
    {
      return "</HEAD>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Header-ben található META adatok megadására használható.
    /// </summary>
    /// <param name="Name">A META adat neve.</param>
    /// <param name="Content">A META adat tartalma.</param>
    /// <returns></returns>
    public static string Meta(string Name,
                              string Content)
    {
      return "<META" + 
                IFNOTNULL(Name,   " name=\'"    + Name    + "\'") + 
                IFNOTNULL(Content," content=\'" + Content + "\'") + 
             ">\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Header-ben található META adatok megadására használható. Ezt a függvényt akkor kell használni ha nem a név adat formátumot akarja használni.
    /// </summary>
    /// <param name="Data">Az adat.</param>
    /// <returns></returns>
    public static string Meta(string Data)
    {
      return "<META" + 
                IFNOTNULL(Data, " " + Data) + 
             ">\n";
    }
    #endregion

    //----------------------------------------------------------------------

    #region BODY Tag
    /// <summary>
    /// Ezzel kell kezdeni a dokumentum testét (BODY).
    /// </summary>
    /// <returns></returns>
    public static string BodyStart()
    {
      return "<BODY>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ezzel kell kezdeni a dokumentum testét (BODY).
    /// </summary>
    /// <param name="BackColor">A dokumentum háttérszíne.</param>
    /// <returns></returns>
    public static string BodyStart(string BackColor)
    {
      return "<BODY" + 
                IFNOTNULL(BackColor, " bgcolor=\'" + BackColor + "\'") + 
             ">\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ezzel kell befejezni a dokumentum testét (BODY).
    /// </summary>
    /// <returns></returns>
    public static string BodyEnd()
    {
      return "</BODY>\n";
    }
    #endregion
    
    #endregion

    //----------------------------------------------------------------------

    #region Font Tags
    /// <summary>
    /// Dölt betus szöveg.
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Italic(string Text)
    {
      return "<I>" + IFNOTNULL(Text, Text) + "</I>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Vastag betus szöveg.
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Bold(string Text)
    {
      return "<B>" + IFNOTNULL(Text, Text) + "</B>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Aláhúzott betus szöveg.
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string UnderLine(string Text)
    {
      return "<U>" + IFNOTNULL(Text, Text) + "</U>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Olyan szöveg melynek style használata nélkül adható meg a betuszine, betumérete és betutipusa.
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <param name="Size">Betuméret.</param>
    /// <param name="Color">Betuszín.</param>
    /// <param name="Family">Betutipus.</param>
    /// <returns></returns>
    public static string Font(string Text,
                              int    Size,
                              string Color,
                              string Family)
    {
      return "<FONT" + 
                IFNOTNULL(Size.ToString(), " size=\'" + Size.ToString() + "\'") + 
                IFNOTNULL(Color, " color=\'" + Color + "\'") + 
                IFNOTNULL(Family," face=\'" + Family + "\'") + 
             ">" + 
                IFNOTNULL(Text, Text) + 
             "</FONT>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Headerek arra használhatóak, hogy dinamikusak legyenek a lap betuméretei. A böngészo "Text Size" beállítása szerint muködnek.A H1(Header 1) a legnagyobb. A H6(Header 6) a legkissebb. 
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Header1(string Text)
    {
      return "<H1>" + IFNOTNULL(Text, Text) + "</H1>\n";    
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Headerek arra használhatóak, hogy dinamikusak legyenek a lap betuméretei. A böngészo "Text Size" beállítása szerint muködnek.A H1(Header 1) a legnagyobb. A H6(Header 6) a legkissebb. 
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Header2(string Text)
    {
      return "<H2>" + IFNOTNULL(Text, Text) + "</H2>\n";    
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Headerek arra használhatóak, hogy dinamikusak legyenek a lap betuméretei. A böngészo "Text Size" beállítása szerint muködnek.A H1(Header 1) a legnagyobb. A H6(Header 6) a legkissebb. 
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Header3(string Text)
    {
      return "<H3>" + IFNOTNULL(Text, Text) + "</H3>\n";    
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Headerek arra használhatóak, hogy dinamikusak legyenek a lap betuméretei. A böngészo "Text Size" beállítása szerint muködnek.A H1(Header 1) a legnagyobb. A H6(Header 6) a legkissebb. 
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Header4(string Text)
    {
      return "<H4>" + IFNOTNULL(Text, Text) + "</H4>\n";    
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Headerek arra használhatóak, hogy dinamikusak legyenek a lap betuméretei. A böngészo "Text Size" beállítása szerint muködnek.A H1(Header 1) a legnagyobb. A H6(Header 6) a legkissebb. 
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Header5(string Text)
    {
      return "<H5>" + IFNOTNULL(Text, Text) + "</H5>\n";    
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A Headerek arra használhatóak, hogy dinamikusak legyenek a lap betuméretei. A böngészo "Text Size" beállítása szerint muködnek.A H1(Header 1) a legnagyobb. A H6(Header 6) a legkissebb. 
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <returns></returns>
    public static string Header6(string Text)
    {
      return "<H6>" + IFNOTNULL(Text, Text) + "</H6>\n";    
    }
    #endregion

    //----------------------------------------------------------------------

    #region Table
    /// <summary>
    /// Ezzel kell kezdeni a táblázatot.
    /// </summary>
    /// <returns></returns>
    public static string TableStart()
    {
      return "<TABLE>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ezzel kell kezdeni a táblázatot.
    /// </summary>
    /// <param name="Border">A keret vastagsága(ha 0 nincs keret).</param>
    /// <param name="CellSpacing">A cellák közötti hely hossza.</param>
    /// <returns></returns>
    public static string TableStart(int Border,
                                    int CellSpacing)
    {
      return "<TABLE" + 
                IFNOTNULL(Border.ToString()," border=\'" + Border.ToString() + "\'") +
                IFNOTNULL(CellSpacing.ToString(), 
                          " cellspacing=\'" + CellSpacing.ToString() + "\'") + 
             ">\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ezzel kell befejezni a táblázatot.
    /// </summary>
    /// <returns></returns>
    public static string TableEnd()
    {
      return "</TABLE>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Új sor kezdete.
    /// </summary>
    /// <returns></returns>
    public static string TRStart()
    {
      return "<TR>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Sor vége.
    /// </summary>
    /// <returns></returns>
    public static string TREnd()
    {
      return "</TR>\n";
    } 

    //----------------------------------------------------------------------

    /// <summary>
    /// Új cella.
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <param name="Class">Style class.</param>
    /// <param name="Width">Cella szélesség(NULL esetén automatikus).</param>
    /// <returns></returns>
    public static string TD(string Text,
                            string Class,
                            int    Width)
    {
      string s;
      s = "<TD" + 
            IFNOTNULL(Class, " class=\'" + Class + "\'") + 
            IFNOTNULL(Width.ToString(), " width=\'" + Width.ToString() + "\'") + 
          ">" +
            IFNOTNULL(Text, Text) + 
          "</TD>\n";
      return s;
    }   

    //----------------------------------------------------------------------

    /// <summary>
    /// Új cella.
    /// </summary>
    /// <param name="Text">A szöveg.</param>
    /// <param name="Class">Style class.</param>
    /// <returns></returns>
    public static string TD(string Text,
                            string Class)
    {
      string s;
      s = "<TD" + 
            IFNOTNULL(Class," class=\'" + Class + "\'") + 
          ">" + 
            IFNOTNULL(Text, Text) + 
          "</TD>\n";
      return s;
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ez a függvény nem része a HTML-nek. DataTable-ból generál Table-t.
    /// </summary>
    /// <param name="TableBorder">Table keret szélessége.</param>
    /// <param name="TableCellSpacing">Table cellái közötti hely hossza.</param>
    /// <param name="CaptionClass">A Caption Style Classa.</param>
    /// <param name="TableClass">A Table Style Classa.</param>
    /// <param name="Data">A DataTable.</param>
    /// <returns></returns>
    public static string GenerateTable(int TableBorder,
                                       int TableCellSpacing,
                                       string    CaptionClass,
                                       string    TableClass,
                                       DataTable Data)
    {
      try
      {
        string ret;
        ret = TableStart(TableBorder,TableCellSpacing);
        ret += TRStart();

        for(int i = 0; i < Data.Columns.Count; i++)
        {
          ret += TD(Data.Columns[i].Caption,CaptionClass);
        }

        ret += TREnd();
        for(int i = 0;i < Data.Rows.Count;i++)
        {
          ret += TRStart();
          for(int i2 = 0; i2 < Data.Rows[i].ItemArray.Length; i2++)
          {
            ret += TD(Data.Rows[i].ItemArray[i2].ToString(),TableClass);                  
          }
          ret += TREnd();
        }

        ret += TableEnd();
        return ret;
      }
      catch(System.Exception e)
      {
        return(GenerateError()+"<h1>Hiba történt a tábla kigenerálása közben!A hibaüzenet:" + e.Message + "</h1>");
      }
    }
    #endregion

    //----------------------------------------------------------------------

    #region Other Tags
    /// <summary>
    /// Új sor.
    /// </summary>
    /// <returns>A HTML elem(ek).</returns>
    public static string BR()
    {
      return "<BR>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// A HTML HR elemének használata (Vonal készítése).
    /// </summary>
    /// <returns>A HTML elem(ek).</returns>
    public static string Line()
    {
      return "<HR>\n";
    }

    //----------------------------------------------------------------------
    
    /// <summary>
    /// A HTML LINK elemének használatával külső style-t tölt be.
    /// </summary>
    /// <param name="src">A file neve.</param>
    /// <param name="ReadFromFile">Ha true kiolvassa a megadott filet, 
    /// igy nincs szükség a css bemásolására a html file mellé.
    /// </param>
    /// <returns>A HTML elem(ek).</returns>
    public static string Style(string src, bool ReadFromFile)
    {
      if(ReadFromFile != true)
      {
        return "<LINK rel=STYLESHEET type=\'text/css\'" + 
                  IFNOTNULL(src, " href=\'" + src + "\'") + 
                  ">\n";
      }
      else
      {
        try
        {
          if(File.Exists(src))
          {
            StreamReader r = new StreamReader(src);
            string ReadedCss = r.ReadToEnd();
            return("<style>" + ReadedCss + "</style>");
          }
          else
          {
            string ret =  string.Format("A <b style=\"color:yellow\">{0}</b> file nem létezik vagy át lett helyezve!", src);
            return(GenerateError() + "<h1>Hiba történt a style kigenerálása közben! A hibaüzenet:" + ret + "</h1>"  + Style(src,false));           
          }
        }
        catch(System.Exception e)
        {
          return(GenerateError()+"<h1>Hiba történt a style kigenerálása közben!A hibaüzenet:" + e.Message + "</h1>" + Style(src,false));
        }
      }
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ha a program generálja a style-t akkor kell használni. A style-t ezzel kell kezdeni.
    /// </summary>
    /// <returns>A HTML elem(ek).</returns>
    public static string StyleStart()
    {
      return "<STYLE>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Ha a program generálja style-t akkor kell használni. A style-t ezzel kell befejezni.
    /// </summary>
    /// <returns>A HTML elem(ek).</returns>
    public static string StyleEnd()
    {
      return "</STYLE>\n";
    }

    //----------------------------------------------------------------------

    /// <summary>
    /// Kép beillesztése a dokumentumba.
    /// </summary>
    /// <param name="src">A file neve.</param>
    /// <param name="Width">A kép szélessége.</param>
    /// <param name="Height">A kép magassága.</param>
    /// <param name="border">A kép keretének szélessége(ha 0 nincs keret).</param>
    /// <param name="Alt">Az a szöveg ami akkor jelenik meg ha a kép nem létezik.</param>
    /// <returns>A HTML elem(ek).</returns>
    public static string Image(string src,
                               int    Width,
                               int    Height,
                               int    Border,
                               string Alt)
    {
      return "<IMG" + 
                IFNOTNULL(src, " src=\'" + src + "\'") + 
                IFNOTNULL(Width.ToString(), " width=\'" + Width + "\'") + 
                IFNOTNULL(Height.ToString(), " height=\'" + Height + "\'") + 
                IFNOTNULL(Border.ToString(), " border=\'" + Border + "\'") + 
                IFNOTNULL(Alt, " alt=\'" + Alt + "\'") + 
              ">";
    }

    //----------------------------------------------------------------------
    
    /// <summary>
    /// BELSŐ FÜGGVÉNY Kérlek ne használd!
    /// </summary>
    /// <returns>-</returns>
    public static string GenerateError()
    {
      return("<style>body{background-color:red;}</style>");
    }

    #endregion
  }
}
