using System;
//using System.Collections;
using System.Diagnostics;
//using System.Threading;
using System.IO;
using System.Text;
using System.Data;
using System.Data.OleDb;
using System.Web;

using JWorldBuilder.HTMLTool;

namespace JWorldBuilder.Logging
{
  //---------------------------------------------------------------------------

  #region enum
  public enum eLoggerMsgType
  {
    none = 0,
    Head,
    Separator,
    StartSeparator,
    StopSeparator,

    Information,
    Info,
    Debug,
    Trace,
    Warning,
    Error,
    Kivtel,
    Exception,
    FNTrace,  // ...           DISABLED !!!
    FNTraceIN,
    FNTraceOUT
  }

  public enum eLoggerMsgStore
  {
    none = 0,
    file,
    DB,
    HTML,
    EventLog
  }
  
  public enum eLogState
  {
    none = 0,   // not available
    opened,
    closed
  }
  #endregion

  //---------------------------------------------------------------------------

  #region LoggerMSGMask
  //  /// <summary>
  //  /// The flag that can store up to 16 flags
  //  /// </summary>
  //  //[CLSCompliant(false)]
  //  public class LoggerMSGMask : BF_16
  //  {
  //    public LoggerMSGMask ()
  //    {
  //      ClearField();
  //    }
  //
  //    //[CLSCompliant(false)]
  //    [Flags]
  //    public enum flag: ushort
  //    {
  //      // 16 flags
  //      empty    = 0x0000,
  //      //     Clear    = 0x0000,
  //
  //      All         = 0x0001,
  //      DisableAll  = 0x0002,
  //      Separator   = 0x0004,  // byte[]
  //      Information = 0x0008,
  //      Warning     = 0x0010,
  //      Error       = 0x0020,  // ChipPy coded
  //      Exception   = 0x0040,
  //      FNTrace     = 0x0080,
  //      //
  ////      Files    = 0x0100,  // Zipped files
  ////      Folder   = 0x0200,  // Zipped dir + sub dirs
  ////      DataSet  = 0x0400,
  ////      Object   = 0x0800,  // any serialized object
  ////      BF_16_mask = 0x1000,
  ////      BF_32_mask = 0x2000
  //      //      F15      = 0x4000,
  //      //      F16      = 0x8000,
  //
  //      //      FAll     = 0xffff //0x7FFF, ha int16
  //    }
  //
  //    #region members
  //    private void SetField(flag p_flag)
  //    {
  //      Mask = (ushort)p_flag;
  //    }
  //
  //    public void SetOn(flag p_flag)
  //    {
  //      Mask |= (ushort)p_flag;
  //    }
  //
  //    public void SetOff(flag p_flag)
  //    {
  //      Mask = Convert.ToUInt16(Mask & ~(ushort)p_flag);
  //    }
  //
  //    public void SetToggle(flag p_flag)
  //    {
  //      Mask ^= (ushort)p_flag;
  //    }
  //
  //    public bool AnyOn (flag p_flag)
  //    {
  //      return (Mask & (ushort)p_flag) != 0;
  //    }
  //
  //    public bool AllOn (flag p_flag)
  //    {
  //      return (Mask & (ushort)p_flag) == (ushort)p_flag;
  //    }
  //
  //    public bool IsOn (flag p_flag)
  //    {
  //      return (Mask & (ushort)p_flag) == (ushort)p_flag;
  //    }
  //
  //    public bool IsEqual(flag p_flag)
  //    {
  //      return Mask == (ushort)p_flag;
  //    }
  //
  //    public override string ToString()
  //      //public new string ToString()
  //    {
  //      return ((flag)Mask).ToString();
  //    }
  //
  //    /// <summary>
  //    /// Convert a decimal value to a Current enum FlagsAttribute value.
  //    /// Method is thread safe
  //    /// </summary>
  //    /// <param name="dec">Valid input: dec between 0,16 </param>
  //    /// <returns></returns>
  //    public static new flag DecimalToFlag(decimal dec)
  //    {
  //      return (flag)BF_16.DecimalToFlag(dec);
  //    }
  //
  //    #endregion
  //  }
  #endregion

  //---------------------------------------------------------------------------
	public class Logger
	{
	  //                     0        1         2         3         4         5         6
	  //                     123456789012345678901234567890123456789012345678901234567890
    string IndentString = "                                                            ";
    int    LineIndent   = 49;
    int    FNCallIndent =  0;
    int    IndentLevel  = -1; // -1: not allowed; 1-60 valid
    bool   _Indent      = false; // autoreset variable

    #region Properties
    //----------------------------------------------------------------------

    private bool _bEnableLog = true;
    public bool bEnableLog
    {
      get
      {
        return _bEnableLog;
      }
      set
      {
        _bEnableLog = value;
      }
    }

    //----------------------------------------------------------------------

    private eLoggerMsgStore _LoggerMsgStore = eLoggerMsgStore.file;
    public eLoggerMsgStore LoggerMsgStore
    {
      get
      {
        return _LoggerMsgStore;
      }
      set
      {
        _LoggerMsgStore = value;
      }
    }

    //----------------------------------------------------------------------

    private string _FulLogFileName  = "";

    private string _defLogName  = "ApplLog";
    private string _defCssName  = "LoggerStyle.css";
    private string _LogFileName = null;
    public string CssFileName   = "LoggerStyle.css";
    public string LogFileName
    {
      get
      {
        return _LogFileName;
      }
      set
      {
        int ind = value.LastIndexOf(".");
        if(ind == -1)
        {
          // add default extension
          switch(LoggerMsgStore)
          {
            case eLoggerMsgStore.file :
              _LogFileName = value + ".log";
              break;

            case eLoggerMsgStore.HTML :
              _LogFileName = value + ".html";
              break;
            default:
              _LogFileName = value;
              break;
          }
        }
        else
        {
          // already keeps extension
          _LogFileName = value;
        }

        if ((_LogFileName != null) && (_FulLogFileName == String.Empty))
        {
          _FulLogFileName = Path.GetFullPath(_LogFileName);
          _LogFileName    = _FulLogFileName;

        }
      }
    }

    //----------------------------------------------------------------------

    private bool _bFullLog = true;
    public bool bFullLog
    {
      get
      {
        return _bFullLog;
      }
      set
      {
        _bFullLog = value;
      }
    }

    //----------------------------------------------------------------------

    private bool _bAutoCallPath = false;
    public bool bAutoCallPath
    {
      get
      {
        return _bAutoCallPath;
      }
      set
      {
        _bAutoCallPath = value;
      }
    }

    //----------------------------------------------------------------------
    #endregion

    #region Constructor
    //----------------------------------------------------------------------

    public Logger(string p_LogFileName)
		{
      if (p_LogFileName == null)
      {
        LogFileName = _defLogName;
      }
      else
      {
        LogFileName = p_LogFileName;
      }
    }

    /// <summary>
    /// CSAK HTML MDBAN HASZNLD!!!!!!!
    /// </summary>
    /// <param name="p_LogFileName"></param>
    /// <param name="p_CssFileName"></param>
    public Logger(string p_LogFileName,
                  string p_CssFileName)
    {
      if (p_LogFileName == null)
      {
        LogFileName = _defLogName;
      }
      else
      {
        LogFileName = p_LogFileName;
      }
      if (p_CssFileName == null)
      {
        CssFileName = _defCssName;
      }
      else
      {
        CssFileName = p_CssFileName;
      }
      LoggerMsgStore = eLoggerMsgStore.HTML;
    }

    //----------------------------------------------------------------------

    public Logger(eLoggerMsgStore p_LoggerMsgStore,
                  string          p_LogFileName)
    {
      LoggerMsgStore = p_LoggerMsgStore;

      if (p_LogFileName == null)
      {
        LogFileName = _defLogName;
      }
      else
      {
        LogFileName = p_LogFileName;
      }
    }

    //----------------------------------------------------------------------

    #endregion

    //======================================================================

    #region Logger implementation

    //----------------------------------------------------------------------
    public eLogState LogState = eLogState.none;

    #region OpenLog
    public void OpenLog()
    {
      if(LogState == eLogState.opened)
      {
        return;
      }
      
      LogState = eLogState.opened;
      Write(1,      // LogID
            100,    // PID
            0,      // Open
            eLoggerMsgType.StartSeparator,
            ">>>>>>>>>>>>>>>>>>>>>>>> Logger Started");
//      Write(0,
//                 eLoggerMsgType.Information,
//                 "======================== Logger Started  ========================");
    }
    #endregion

    //----------------------------------------------------------------------

    #region CloseLog
    public void CloseLog()
    {
      if(LogState == eLogState.opened)
      {
        LogState = eLogState.closed;
  //      Write(-1,
  //                 eLoggerMsgType.Information,
  //                 "======================== Logger Ended =========================");
        Write(1,      // LogID
              100,    // PID
              -1,     // Close
              eLoggerMsgType.StopSeparator,
              "<<<<<<<<<<<<<<<<<<<<<<<< Logger Ended");
      }
      return;     
    }
    #endregion

    //----------------------------------------------------------------------

    #region Logger interface

    #region WriteInfo

    //----------------------------------------------------------------------

    private bool _bEnableInfo = true;
    public bool bEnableInfo
    {
      get
      {
        return _bEnableInfo;
      }
      set
      {
        _bEnableInfo = value;
      }
    }

    //----------------------------------------------------------------------

    public bool WriteInfoIf(bool   p_condition,
                            string p_Message)
    {
      if(!p_condition || !_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Info,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteInfoIf(bool   p_condition,
                            int    p_MsgID,
                            string p_Message)
    {
      if(!p_condition || !_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Info,
                   p_Message);
    }

    //----------------------------------------------------------------------
     public bool WriteInfo(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Info,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteInfo(int    p_MsgID,
                          string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Info,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    #region WriteWarning

    //----------------------------------------------------------------------

    private bool _bEnableWarning = true;
    public bool bEnableWarning
    {
      get
      {
        return _bEnableWarning;
      }
      set
      {
        _bEnableWarning = value;
      }
    }

    //----------------------------------------------------------------------

    public bool WriteWarningIf(bool   p_condition,
                               string p_Message)
    {
      if(!p_condition || !_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Warning,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteWarningIf(bool   p_condition,
                               int    p_MsgID,
                               string p_Message)
    {
      if(!p_condition || !_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Warning,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteWarning(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Warning,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteWarning(int    p_MsgID,
                             string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Warning,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    #region WriteError

    //----------------------------------------------------------------------

    private bool _bEnableError = true;
    public bool bEnableError
    {
      get
      {
        return _bEnableError;
      }
      set
      {
        _bEnableError = value;
      }
    }

    //----------------------------------------------------------------------

    public bool WriteErrorIf(bool   p_condition,
                             string p_Message)
    {
      if(!p_condition || !_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Error,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteErrorIf(bool   p_condition,
                             int    p_MsgID,
                             string p_Message)
    {
      if(!p_condition || !_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Error,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteError(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Error,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteError(int    p_MsgID,
                           string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Error,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    #region WriteException

    //----------------------------------------------------------------------

    private bool _bEnableException = true;
    public bool bEnableException
    {
      get
      {
        return _bEnableException;
      }
      set
      {
        _bEnableException = value;
      }
    }

    //----------------------------------------------------------------------

    public bool WriteExceptionIf(bool   p_condition,
                                 string p_Message)
    {
      if(!p_condition || !_bEnableLog || !_bEnableException)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Exception,    //Exception,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteExceptionIf(bool   p_condition,
                                 int    p_MsgID,
                                 string p_Message)
    {
      if(!p_condition || !_bEnableLog || !_bEnableException)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Exception,    //Exception,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteException(string p_Message)
    {
      if(!_bEnableLog || !_bEnableException)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.Exception,    //Exception,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteException(int    p_MsgID,
                               string p_Message)
    {
      if(!_bEnableLog || !_bEnableException)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.Exception,    //Exception,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    //----------------------------------------------------------------------

//    [ Conditional("LOGGER") ]
    #region Info...
    #region InfoIf
    // -----------------------------------

    public void InfoIf(bool p_condition)
    {
      if(p_condition)
      {
        InfoInt(0,
                null);
      }
    }

    // -----------------------------------

    public void InfoIf(bool   p_condition,
                       string p_Message)
    {
      if(p_condition)
      {
        InfoInt(0,
                p_Message);
      }
    }

    // -----------------------------------

    public void InfoIf(bool p_condition,
                       int  p_MsgID)
    {
      if(p_condition)
      {
        InfoInt(p_MsgID,
                null);
      }
    }

    // -----------------------------------

    public void InfoIf(bool   p_condition,
                       int    p_MsgID,
                       string p_Message)
    {
      if(p_condition)
      {
        InfoInt(p_MsgID,
                p_Message);
      }
    }

    // -----------------------------------
    #endregion
    #region Info
    // -----------------------------------

    public void Info()
    {
      InfoInt(0,
              null);
    }

    // -----------------------------------

    public void Info(string p_Message)
    {
      InfoInt(0,
              p_Message);
    }

    // -----------------------------------

    public void Info(int    p_MsgID)
    {
      InfoInt(p_MsgID,
              null);
    }

    // -----------------------------------

    public void Info(int    p_MsgID,
                     string p_Message)
    {
      InfoInt(p_MsgID,
              p_Message);
    }

    // -----------------------------------

    #endregion

    // ===================================

    //    [ Conditional("LOGGER") ]
    #region InfoInt
    private void InfoInt(int    p_MsgID,
                         string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      if (bAutoCallPath)
      {
        _DisableNL = ((p_Message == null) ||
          (LoggerMsgStore == eLoggerMsgStore.file));
        if(_bFullLog)
        {
          CallPath = GetFullMemberInfoWithLineNumberNL(sf);
        }
        else
        {
          CallPath = GetMemberInfoNL(sf);
        }
        Write(p_MsgID,
              eLoggerMsgType.Info,
              CallPath +
              HTML.IFNOTNULL(p_Message, ": [" + p_Message + "]"));
      }
      else
      {
        Write(p_MsgID,
              eLoggerMsgType.Info,
              HTML.IFNOTNULL(p_Message, "[" + p_Message + "]"));
      }
    }
    #endregion
    #endregion

    //----------------------------------------------------------------------

//    [ Conditional("DEBUG") ]
//    [ Conditional("TRACE") ]
    #region Condition trace  [ Conditional("DEBUG") ], "TRACE"

    //----------------------------------------------------------------------

    #region FNTrace...
    #region FNTraceIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTraceIf(bool p_condition)
    {
      if(p_condition)
      {
        FNTraceInt(0,
                   null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTraceIf(bool   p_condition,
                          string p_Message)
    {
      if(p_condition)
      {
        FNTraceInt(0,
                   p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTraceIf(bool p_condition,
                          int  p_MsgID)
    {
      if(p_condition)
      {
        FNTraceInt(p_MsgID,
                   null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTraceIf(bool   p_condition,
                          int    p_MsgID,
                          string p_Message)
    {
      if(p_condition)
      {
        FNTraceInt(p_MsgID,
                   p_Message);
      }
    }

    // -----------------------------------
    #endregion
    #region FNTrace
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTrace()
    {
      FNTraceInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTrace(string p_Message)
    {
      FNTraceInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTrace(int    p_MsgID)
    {
      FNTraceInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void FNTrace(int    p_MsgID,
                        string p_Message)
    {
      FNTraceInt(p_MsgID, p_Message);
    }
    // -----------------------------------
    #endregion

    // ===================================

    [ Conditional("DEBUG") ]
    #region FNTraceInt
    public void FNTraceInt(int    p_MsgID,
                           string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      _DisableNL = ((p_Message == null) ||
                    (LoggerMsgStore == eLoggerMsgStore.file));
      if(_bFullLog)
      {
        CallPath = GetFullMemberInfoWithLineNumberNL(sf);
      }
      else
      {
        CallPath = GetMemberInfoWithLineNumberNL(sf);
      }

      Write(p_MsgID,
            eLoggerMsgType.FNTrace,
            CallPath +
            HTML.IFNOTNULL(p_Message, ": [" + p_Message + "]"));
    }
    #endregion
    #endregion

    //----------------------------------------------------------------------

    #region Enter...
    #region EnterIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void EnterIf(bool p_condition)
    {
      if(p_condition)
      {
        EnterInt(0, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void EnterIf(bool   p_condition,
                        string p_Message)
    {
      if(p_condition)
      {
        EnterInt(0, p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void EnterIf(bool p_condition,
                        int  p_MsgID)
    {
      if(p_condition)
      {
        EnterInt(p_MsgID, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void EnterIf(bool   p_condition,
                        int    p_MsgID,
                        string p_Message)
    {
      if(p_condition)
      {
        EnterInt(p_MsgID, p_Message);
      }
    }

    // -----------------------------------
    #endregion
    #region Enter
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Enter()
    {
      EnterInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Enter(string p_Message)
    {
      EnterInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Enter(int    p_MsgID)
    {
      EnterInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Enter(int    p_MsgID,
                      string p_Message)
    {
      EnterInt(p_MsgID, p_Message);
    }
    // -----------------------------------
    #endregion

    // ===================================

    [ Conditional("DEBUG") ]
    #region EnterInt
    public void EnterInt(int    p_MsgID,
                         string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      _DisableNL = ((p_Message == null) ||
                    (LoggerMsgStore == eLoggerMsgStore.file));
      if(_bFullLog)
      {
        CallPath = GetFullMemberInfoWithLineNumberNL(sf);
      }
      else
      {
        CallPath = GetMemberInfoWithLineNumberNL(sf);
      }

      Write(p_MsgID,
            eLoggerMsgType.FNTraceIN,
            CallPath +
            HTML.IFNOTNULL(p_Message, ": [" + p_Message + "]"));
    }
    #endregion
    #endregion

    //----------------------------------------------------------------------

    #region Leave...
    #region LeaveIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void LeaveIf(bool p_condition)
    {
      if(p_condition)
      {
        LeaveInt(0, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void LeaveIf(bool   p_condition,
                        string p_Message)
    {
      if(p_condition)
      {
        LeaveInt(0, p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void LeaveIf(bool p_condition,
                        int  p_MsgID)
    {
      if(p_condition)
      {
        LeaveInt(p_MsgID, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void LeaveIf(bool   p_condition,
                        int    p_MsgID,
                        string p_Message)
    {
      if(p_condition)
      {
        LeaveInt(p_MsgID, p_Message);
      }
    }
    // -----------------------------------
    #endregion
    #region Leave
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Leave()
    {
      LeaveInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Leave(string p_Message)
    {
      LeaveInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Leave(int    p_MsgID)
    {
      LeaveInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Leave(int    p_MsgID,
                      string p_Message)
    {
      LeaveInt(p_MsgID, p_Message);
    }
    // -----------------------------------
    #endregion

    // ===================================

    [ Conditional("DEBUG") ]
    #region LeaveInt
    public void LeaveInt(int    p_MsgID,
                         string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      _DisableNL = ((p_Message == null) ||
                    (LoggerMsgStore == eLoggerMsgStore.file));
      if(_bFullLog)
      {
        CallPath = GetFullMemberInfoWithLineNumberNL(sf);
      }
      else
      {
        CallPath = GetMemberInfoWithLineNumberNL(sf);
      }

      Write(p_MsgID,
            eLoggerMsgType.FNTraceOUT,
            CallPath +
            HTML.IFNOTNULL(p_Message, ": [" + p_Message + "]"));
    }
    #endregion
    #endregion

    //----------------------------------------------------------------------

    #region Trace...
    #region TraceIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void TraceIf(bool p_condition)
    {
      if(p_condition)
      {
        TraceInt(0, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void TraceIf(bool   p_condition,
                        string p_Message)
    {
      if(p_condition)
      {
        TraceInt(0, p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void TraceIf(bool p_condition,
                        int  p_MsgID)
    {
      if(p_condition)
      {
        TraceInt(p_MsgID, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void TraceIf(bool   p_condition,
                        int    p_MsgID,
                        string p_Message)
    {
      if(p_condition)
      {
        TraceInt(p_MsgID, p_Message);
      }
    }
    // -----------------------------------
    #endregion
    #region Trace
    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Trace()
    {
      TraceInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Trace(string p_Message)
    {
      TraceInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Trace(int    p_MsgID)
    {
      TraceInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Trace(int    p_MsgID,
                      string p_Message)
    {
      TraceInt(p_MsgID, p_Message);
    }
    // -----------------------------------

    #endregion

    // ===================================

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    #region TraceInt
    public void TraceInt(int    p_MsgID,
                         string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      if (bAutoCallPath)
      {
        _DisableNL = ((p_Message == null) ||
                      (LoggerMsgStore == eLoggerMsgStore.file));
        if(_bFullLog)
        {
          CallPath = GetFullMemberInfoWithLineNumberNL(sf);
        }
        else
        {
          CallPath = GetMemberInfoNL(sf);
        }

        Write(p_MsgID,
              eLoggerMsgType.Trace,
              CallPath +
              HTML.IFNOTNULL(p_Message, ": [" + p_Message + "]"));
      }
      else
      {
        Write(p_MsgID,
              eLoggerMsgType.Trace,
              HTML.IFNOTNULL(p_Message, "[" + p_Message + "]"));
      }
      #endregion
    }
    #endregion

    //----------------------------------------------------------------------

    #region Debug...
    #region DebugIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void DebugIf(bool p_condition)
    {
      if(p_condition)
      {
        DebugInt(0, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void DebugIf(bool   p_condition,
                        string p_Message)
    {
      if(p_condition)
      {
        DebugInt(0, p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void DebugIf(bool p_condition,
                        int  p_MsgID)
    {
      if(p_condition)
      {
        DebugInt(p_MsgID, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void DebugIf(bool   p_condition,
                        int    p_MsgID,
                        string p_Message)
    {
      if(p_condition)
      {
        DebugInt(p_MsgID, p_Message);
      }
    }
    // -----------------------------------
    #endregion
    #region Debug
    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Debug()
    {
      DebugInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Debug(string p_Message)
    {
      DebugInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Debug(int    p_MsgID)
    {
      DebugInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    public void Debug(int    p_MsgID,
                      string p_Message)
    {
      DebugInt(p_MsgID, p_Message);
    }
    // -----------------------------------
    #endregion

    // ===================================

    [ Conditional("DEBUG") ]
    #region DebugInt
    private void DebugInt(int    p_MsgID,
                          string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      if (bAutoCallPath)
      {
        _DisableNL = ((p_Message == null) ||
                      (LoggerMsgStore == eLoggerMsgStore.file));
        if(_bFullLog)
        {
          CallPath = GetFullMemberInfoWithLineNumberNL(sf);
        }
        else
        {
          CallPath = GetMemberInfoNL(sf);
        }
        Write(p_MsgID,
              eLoggerMsgType.Debug,
              CallPath +
              HTML.IFNOTNULL(p_Message, ": [" + p_Message + "]"));
      }
      else
      {
        Write(p_MsgID,
              eLoggerMsgType.Debug,
              HTML.IFNOTNULL(p_Message, "[" + p_Message + "]"));
      }
      #endregion

    }
    #endregion

    //----------------------------------------------------------------------

    #region Error...
    #region ErrorIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ErrorIf(bool p_condition)
    {
      if(p_condition)
      {
        ErrorInt(0, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ErrorIf(bool   p_condition,
                        string p_Message)
    {
      if(p_condition)
      {
        ErrorInt(0, p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ErrorIf(bool p_condition,
                        int  p_MsgID)
    {
      if(p_condition)
      {
        ErrorInt(p_MsgID, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ErrorIf(bool   p_condition,
                        int    p_MsgID,
                        string p_Message)
    {
      if(p_condition)
      {
        ErrorInt(p_MsgID, p_Message);
      }
    }
    // -----------------------------------
    #endregion
    #region Error
    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Error()
    {
      ErrorInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Error(string p_Message)
    {
      ErrorInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Error(int    p_MsgID)
    {
      ErrorInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Error(int    p_MsgID,
                      string p_Message)
    {
      ErrorInt(p_MsgID, p_Message);
    }

    // -----------------------------------
    #endregion

    // ===================================

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    #region ErrorInt
    private void ErrorInt(int    p_MsgID,
                          string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      _DisableNL = (p_Message == null);
//      _Indent    = (LoggerMsgStore == eLoggerMsgStore.file);

      if(_bFullLog)
      {
        CallPath = GetFullMemberInfoWithLineNumberNL(sf);
      }
      else
      {
        CallPath = GetMemberInfoWithLineNumberNL(sf);
      }

      Write(p_MsgID,
            eLoggerMsgType.Error,
            CallPath +
            HTML.IFNOTNULL(p_Message, "[" + p_Message + "]"));
    }
    #endregion
    #endregion

    //----------------------------------------------------------------------

    #region Exception...
    #region ExceptionIf
    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ExceptionIf(bool p_condition)
    {
      if(p_condition)
      {
        ExceptionInt(0, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ExceptionIf(bool   p_condition,
                            string p_Message)
    {
      if(p_condition)
      {
        ExceptionInt(0, p_Message);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ExceptionIf(bool p_condition,
                            int  p_MsgID)
    {
      if(p_condition)
      {
        ExceptionInt(p_MsgID, null);
      }
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void ExceptionIf(bool   p_condition,
                            int    p_MsgID,
                            string p_Message)
    {
      if(p_condition)
      {
        ExceptionInt(p_MsgID, p_Message);
      }
    }
    // -----------------------------------
    #endregion
    #region Exception
    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Exception()
    {
      ExceptionInt(0, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Exception(string p_Message)
    {
      ExceptionInt(0, p_Message);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Exception(int    p_MsgID)
    {
      ExceptionInt(p_MsgID, null);
    }

    // -----------------------------------

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    public void Exception(int    p_MsgID,
                          string p_Message)
    {
      ExceptionInt(p_MsgID, p_Message);
    }

    // -----------------------------------
    #endregion
      // ===================================

    [ Conditional("DEBUG") ]
    [ Conditional("TRACE") ]
    #region ExceptionInt
    private void ExceptionInt(int    p_MsgID,
                              string p_Message)
    {
      int nFrameInd = 2;  //(p_Message == null) ? 2 : 1;    // Correct the indirect call !!!
      #region t
      StackTrace st = new StackTrace();
      string _calledMethod  = st.GetFrame(1).GetFileName();
      string _callingMethod = new StackTrace().GetFrame(1).GetMethod().Name;
      string s = st.GetFrame(nFrameInd).GetMethod().ToString();
      #endregion
      StackFrame sf = new StackFrame(nFrameInd,true); // init with the caller info
      string CallPath = null;

      _DisableNL = (p_Message == null);
//      _Indent    = (LoggerMsgStore == eLoggerMsgStore.file);

      if(_bFullLog)
      {
        CallPath = GetFullMemberInfoWithLineNumberNL(sf);
      }
      else
      {
        CallPath = GetMemberInfoWithLineNumberNL(sf);
      }

      Write(p_MsgID,
            eLoggerMsgType.Exception,
            CallPath +                    // st.GetFrame(nFrameInd).GetMethod().Name +
            HTML.IFNOTNULL(p_Message, "[" + p_Message + "]"));
    }
    #endregion
    #endregion

    //----------------------------------------------------------------------

    #region GetMemberInfo...

    //---------------------------------------------------
    // SHORT format
    //---------------------------------------------------

    private string GetMemberInfo(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           false,   //p_full,
                           false,   //p_LineNumber,
                           false    //p_NewLine)
        );
    }

    //---------------------------------------------------

    private string GetMemberInfoNL(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           false,   //p_full,
                           false,   //p_LineNumber,
                           true     //p_NewLine)
                           );
    }

    //---------------------------------------------------

    private string GetMemberInfoWithLineNumber(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           false,   //p_full,
                           true,    //p_LineNumber,
                           false    //p_NewLine)
                          );
    }

    //---------------------------------------------------

    private string GetMemberInfoWithLineNumberNL(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           false,   //p_full,
                           true,    //p_LineNumber,
                           true     //p_NewLine)
                          );
    }

    //---------------------------------------------------
    // FULL format
    //---------------------------------------------------


    private string GetFullMemberInfo(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           true,   //p_full,
                           false,   //p_LineNumber,
                           false    //p_NewLine)
                          );
    }

    //---------------------------------------------------

    private string GetFullMemberInfoNL(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           true,    //p_full,
                           false,   //p_LineNumber,
                           true     //p_NewLine)
                          );
    }

    //---------------------------------------------------

    private string GetFullMemberInfoWithLineNumber(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           true,   //p_full,
                           true,    //p_LineNumber,
                           false    //p_NewLine)
                          );
    }

    //---------------------------------------------------

    private string GetFullMemberInfoWithLineNumberNL(StackFrame p_sf)
    {
      return GetMemberInfo(p_sf ,
                           true,    //p_full,
                           true,    //p_LineNumber,
                           true     //p_NewLine)
                          );
    }

    //---------------------------------------------------
    // base implementation
    //---------------------------------------------------

    // this is a special variable, Disable the NewLine on the end of the generated
    // string, and automatically will be reset the
    private bool _DisableNL = false;

    private string GetMemberInfo(StackFrame p_sf,
                                 bool       p_full,
                                 bool       p_LineNumber,
                                 bool       p_NewLine)
    {

// format like this:
//at FTPStore.frmFTPTestClient.SetMore() in d:\cybermouse\__myims\test\___imcryptoserver\ftp\ftpstore\frmftptestclient.cs:line 3030
//   at FTPStore.frmFTPTestClient.btnMore_Click(Object sender, EventArgs e) in d:\cybermouse\__myims\test\___imcryptoserver\ftp\ftpstore\frmftptestclient.cs:line 3008
//   at System.Windows.Forms.Control.OnClick(EventArgs e
      string RealFileName   = p_sf.GetFileName();
      string FullFileName   = Path.GetFileName( RealFileName);

      StringBuilder sb = new StringBuilder(512);

      if(p_full)
      {
        string fullMethodName = p_sf.GetMethod().ToString();
        sb.Append ( "at "   + fullMethodName +
                    HTML.IFNOTNULL(RealFileName, " in "  + RealFileName));
      }
      else
      {
        string FileName       = Path.GetFileNameWithoutExtension( RealFileName );
        string MethodName     = p_sf.GetMethod().Name;
//        sb.Append (FileName + "." + MethodName);
        sb.Append ("'" + HTML.IFNOTNULL(FileName, FileName + ".") + MethodName + "'");
      }

      if(p_LineNumber)
      {
        int    LineNumber     = p_sf.GetFileLineNumber();
        if(LineNumber > 0)
        {
          sb.Append (":line " + LineNumber);
        }
      }

      if(!_DisableNL && p_NewLine)
      {
        sb.Append (Environment.NewLine);
        if(_Indent)
        {
          sb.Append (IndentString.Substring(0, LineIndent));
        }
      }

      _DisableNL = false;
      _Indent    = false;
      string CallPath = sb.ToString();

      return CallPath;
    }
    #endregion

    //----------------------------------------------------------------------

    #endregion

    //----------------------------------------------------------------------

    #region WriteFNTrace ...           DISABLED !!!

//    //----------------------------------------------------------------------

    private bool _bEnableFNTrace = true;
    public bool bEnableFNTrace
    {
      get
      {
        return (_bEnableFNTraceIN || _bEnableFNTraceOUT);
      }
      set
      {
        _bEnableFNTrace    = value;
        _bEnableFNTraceIN  = value;
        _bEnableFNTraceOUT = value;
      }
    }

    //----------------------------------------------------------------------

    public bool WriteFNTrace(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.FNTrace,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteFNTrace(int    p_MsgID,
                             string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.FNTrace,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    #region WriteFNTraceIN

    //----------------------------------------------------------------------

    private bool _bEnableFNTraceIN = true;
    public bool bEnableFNTraceIN
    {
      get
      {
        return _bEnableFNTraceIN;
      }
      set
      {
        _bEnableFNTraceIN = value;
        _bEnableFNTrace   = (_bEnableFNTraceIN || _bEnableFNTraceOUT);
      }
    }

    //----------------------------------------------------------------------

    public bool WriteFNTraceIN(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.FNTraceIN,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteFNTraceIN(int    p_MsgID,
                               string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.FNTraceIN,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    #region WriteFNTraceOUT

    //----------------------------------------------------------------------

    private bool _bEnableFNTraceOUT = true;
    public bool bEnableFNTraceOUT
    {
      get
      {
        return _bEnableFNTraceOUT;
      }
      set
      {
        _bEnableFNTraceOUT = value;
        _bEnableFNTrace    = (_bEnableFNTraceIN || _bEnableFNTraceOUT);
      }
    }

    //----------------------------------------------------------------------

    public bool WriteFNTraceOUT(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.FNTraceOUT,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool WriteFNTraceOUT(int    p_MsgID,
                                string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.FNTraceOUT,
                   p_Message);
    }

    //----------------------------------------------------------------------
    #endregion

    #endregion

    //----------------------------------------------------------------------

    #region LogTimeToString
    public string LogTimeToString (DateTime p_DT)
    {
      DateTime DT = p_DT;   //DateTime.Now;

      string sDate = DT.Year + "/" +
                     DT.Month.ToString("0#")  + "/" +
                     DT.Day.ToString("0#")    + ", " +
                     DT.Hour.ToString("0#")   + ":" +
                     DT.Minute.ToString("0#") + ":" +
                     DT.Second.ToString("0#") + "." +
                     DT.Millisecond.ToString("00#");
      return sDate;
    }
    #endregion

    //----------------------------------------------------------------------

    #region IsEnabled
    /// <summary>
    ///
    /// </summary>
    /// <param name="p_MsgType"></param>
    /// <returns>true:  the log is enabled</returns>
    /// <returns>false: the log is not enabled</returns>
    public bool IsEnabled(eLoggerMsgType p_MsgType)
    {
      bool ok = true; // enabled

      if(!_bEnableLog)
      {
        return false; // disabled
      }

      switch(p_MsgType)
      {
        case eLoggerMsgType.Information:
        case eLoggerMsgType.Info:
          ok = _bEnableInfo;
          break;

        case eLoggerMsgType.Warning:
            ok = _bEnableWarning;
          break;

        case eLoggerMsgType.Error:
            ok = _bEnableError;
          break;

        case eLoggerMsgType.Kivtel:
        case eLoggerMsgType.Exception:
            ok = _bEnableException;
          break;

        case eLoggerMsgType.FNTrace:
            ok = _bEnableFNTrace;
          break;

        case eLoggerMsgType.FNTraceIN:
            ok = _bEnableFNTraceIN;
          break;

        case eLoggerMsgType.FNTraceOUT:
            ok = _bEnableFNTraceOUT;
          break;

        default:
            // all others are enabled!!! not disabled!
          break;
      }

      return ok;;
    }
    #endregion

    //----------------------------------------------------------------------

    #region Write
    //----------------------------------------------------------------------

    public bool Write(string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(0,
                   eLoggerMsgType.none,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool Write(int    p_MsgID,
                      string p_Message)
    {
      if(!_bEnableLog)
      {
        return true;
      }

      return Write(p_MsgID,
                   eLoggerMsgType.none,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool Write(int            p_MsgID,
                      eLoggerMsgType p_MsgType,
                      string         p_Message)
    {
      if(!IsEnabled(p_MsgType))
      {
        return true;
      }

      return Write(-1,    //p_LogID,
                   -1,    //p_PID,
                   p_MsgID,
                   p_MsgType,
                   p_Message);
    }

    //----------------------------------------------------------------------

    public bool Write(int            p_LogID, // sub log ID, created by OpenLog()
                      int            p_PID,   // task ID
                      int            p_MsgID,
                      eLoggerMsgType p_MsgType,
                      string         p_Message)
    {
      bool ok = false;

      if(!IsEnabled(p_MsgType))
      {
        return true;
      }

      switch(LoggerMsgStore)
      {
        case eLoggerMsgStore.file:
                                    return WriteToFile(p_MsgID,
                                                       p_MsgType,
                                                       p_Message);
                                    //break;

        case eLoggerMsgStore.HTML:
                                    return WriteToHTML(p_MsgID,
                                                       p_MsgType,
                                                       p_Message);
                                    //break;
        case eLoggerMsgStore.DB:
                                    return WriteToDB( p_LogID, // sub log ID
                                                      p_PID,   // task ID
                                                      p_MsgID,
                                                      p_MsgType,
                                                      p_Message);
                                    //break;
        default:
                                    return ok;
                                    //break;
      }
    }

    #endregion

    //----------------------------------------------------------------------

    #region WriteToFile
    private bool WriteToFile(int            p_MsgID,
                             eLoggerMsgType p_MsgType,
                             string         p_Message)
    {
      bool ok = true;
      //========================================================================

      lock(typeof(Logger))
      {
        #region Locked part
        StreamWriter sw = File.AppendText( LogFileName );

        //========================================================================

        // !!!!!
        p_Message = p_Message.Replace("\r\n",
                                      "\r\n" + IndentString.Substring(0, LineIndent));
        try
        {
          switch(p_MsgType)
          {
            case eLoggerMsgType.StartSeparator :
              #region StartSeparator
              //========================================================================

              sw.Write(Environment.NewLine);
              sw.Write(">>>> ");
              sw.Write(LogTimeToString(DateTime.Now) + " >>>>>>>>");
              sw.Write(p_Message);

              sw.Write(Environment.NewLine);
              //========================================================================
              #endregion
              break;

            case eLoggerMsgType.StopSeparator :
              #region StopSeparator
              //========================================================================

              sw.Write("<<<< ");
              sw.Write(LogTimeToString(DateTime.Now) + " <<<<<<<<");
              sw.Write(p_Message);

              sw.Write(Environment.NewLine);

              //========================================================================
              #endregion
              break;

            default:
              #region default
              //=== ID =================================================================

              string sID    = "-------- ";  // 8 + 1
              if(p_MsgID <= 0)
              {
                sw.Write(sID);
              }
              else
              {
                sw.Write(p_MsgID.ToString("000000#") + ", ");
              }

              //=== Date ===============================================================

              sw.Write(LogTimeToString(DateTime.Now) + " - ");

              //=== MSG type ===========================================================
              //=== Message ============================================================

              string sFNIN    = "==>\t";
              string sFNOUT   = "\t<==";
              string sMsgType = "";

              if(p_MsgType == eLoggerMsgType.none)
              {
                // add Msg type + message to the log
                sw.Write(p_Message);
              }
              else
              {
                switch(p_MsgType)
                {
                  case eLoggerMsgType.FNTraceIN :
                    sMsgType = sFNIN;
                    break;

                  case eLoggerMsgType.FNTraceOUT :
                    sMsgType = sFNOUT;
                    break;

                  default :
                    sMsgType = p_MsgType.ToString();
                    break;
                }

                // add Msg type + message to the log
                sw.Write("\t" + sMsgType + "\t[" + p_Message + "]");
              }

              sw.Write(Environment.NewLine);

              //========================================================================
              #endregion
              break;
          }
          ok = true;
        }
        catch
        {
          ok = false;
        }
        finally
        {
          sw.Close();
        }
        #endregion
      }

      return ok;
    }
    #endregion

    //----------------------------------------------------------------------

    #region WriteToDB
    // demo sql format !!!
    //http://www.csharphelp.com/archives/archive37.html
    private bool WriteToDB(int            p_LogID, // sub log ID, created by OpenLog()
                           int            p_PID,   // task ID
                           int            p_MsgID,
                           eLoggerMsgType p_MsgType,
                           string         p_Message)
    {
      bool ok;
      string sConnect = "Provider=Microsoft.Jet.OLEDB.4.0;User Id=;Password=;Data Source=";
      OleDbConnection conn = new OleDbConnection(sConnect + LogFileName + ";");
      OleDbCommand    comm = new OleDbCommand();
      try
      {
        DateTime DT = DateTime.Now;
        conn.Open();
        comm.Connection  = conn;
//        String EncodeString = System.Web.Server.UrlEncode(p_Message);
        string EncodeString = p_Message.Replace("\"","\\\"");
               EncodeString = EncodeString.Replace("'","^");

    #region x
//        string command2 = String.Format("insert into LogTable(MsgID, [Date], MsgType, Message) values({0},'{1}','{2}','{3}')",
//                                        p_MsgID.ToString(),
//                                        DT,
//                                        p_MsgType,
//                                        p_Message);
#endregion
        comm.CommandText = "INSERT INTO LogTable " +
                           "(LogID, PID, MsgID, [Date], MsgType, Message) VALUES ('" +
                           p_LogID.ToString() + "', '" +
                           p_PID.ToString()   + "', '" +
                           p_MsgID.ToString() + "', '" +
                           DT + "', '" +                     //.ToString()
                           p_MsgType + "', '" +              //.ToString()
//                           p_Message + "')";
                           EncodeString + "')";

        comm.ExecuteNonQuery();
        ok = true;
      }
      catch(System.IO.IOException e)
      {
        ok = false;
      }
      finally
      {
        comm.Dispose();
        conn.Close();
      }
      return ok;
    }
    #endregion

    //----------------------------------------------------------------------

    #region WriteSeparatorLine
    private void WriteHTMLSeparatorLine (StreamWriter   p_sw,
                                         int            p_MsgID,
                                         eLoggerMsgType p_MsgType,
                                         eLoggerMsgType p_FormatMsgType,
                                         string         p_Message)
    {
      string sMsgType  = "";
      string sMsgStyle = p_FormatMsgType.ToString();
      bool   bBold     = false;

      p_sw.Write(HTML.TRStart());

      //=== ID =================================================================

      switch(p_MsgType)
      {
        case eLoggerMsgType.StartSeparator :
          // StartSeparator
          bBold = true;
          sMsgType = ">>>>>>>>";
          p_sw.Write(HTML.TD(sMsgType, sMsgStyle));
          break;

        case eLoggerMsgType.StopSeparator :
          // StopSeparator
          bBold = true;
          sMsgType = "<<<<<<<<";
          p_sw.Write(HTML.TD(sMsgType, sMsgStyle));
          break;

        default:
          string sID = "--------";

          if(p_MsgID <= 0)
          {
            p_sw.Write(HTML.TD(sID, sMsgStyle));
          }
          else
          {
            p_sw.Write(HTML.TD(p_MsgID.ToString("000000#"), sMsgStyle));
          }
          break;
      }

      //=== Date ===============================================================

      p_sw.Write(HTML.TD(LogTimeToString(DateTime.Now), sMsgStyle));

      //=== MSG type ===========================================================
      //=== Message ============================================================

      // add Msg type + message to the log
      if(bBold)
      {
        p_sw.Write(HTML.TD(HTML.Bold(sMsgType),  sMsgStyle + "T"));
        p_sw.Write(HTML.TD(HTML.Bold(p_Message), sMsgStyle + "T"));
      }
      else
      {
        p_sw.Write(HTML.TD(sMsgType,  sMsgStyle + "T"));
        p_sw.Write(HTML.TD(p_Message, sMsgStyle + "T"));
      }

      //========================================================================

      p_sw.Write(HTML.TREnd());

//      p_sw.Write(Environment.NewLine);

      //========================================================================
    }
    #endregion

    //----------------------------------------------------------------------

    #region WriteToHTML
    private bool WriteToHTML(int            p_MsgID,
                             eLoggerMsgType p_MsgType,
                             string         p_Message)
    {
      bool   ok;
      string sMsgStyle = p_MsgType.ToString();

      //========================================================================

      lock(typeof(Logger))
      {
        #region Locked part
        StreamWriter sw;
        if(File.Exists( LogFileName ))
        {
          sw = File.AppendText( LogFileName );
        }
        else
        {
          string fullPath = Path.GetDirectoryName(LogFileName);

          // Determine whether the directory exists.
          if (!Directory.Exists(fullPath))
          {
            // Create the directory it does not exist.
            Directory.CreateDirectory(fullPath);
          }

          sw = File.AppendText( LogFileName );
          //sw.Write(HTML.Style("LoggerStyle.css"));
          sw.Write(HTML.Style(CssFileName,
                              true));
        }

        //========================================================================

        // !!!!!
        p_Message = p_Message.Replace("\r\n", "<br>");

        try
        {
          switch(p_MsgType)
          {
            case eLoggerMsgType.StartSeparator :
              #region StartSeparator
              //========================================================================

              sw.Write(HTML.TableStart(1,0));

              sw.Write(HTML.TRStart());
                sw.Write(HTML.TD("ID",      eLoggerMsgType.Head.ToString()));
                sw.Write(HTML.TD("Date",    eLoggerMsgType.Head.ToString()));
                sw.Write(HTML.TD("Type",    eLoggerMsgType.Head.ToString()));
                sw.Write(HTML.TD("Message", eLoggerMsgType.Head.ToString()));
              sw.Write(HTML.TREnd());

              WriteHTMLSeparatorLine(sw,
                                     p_MsgID,
                                     p_MsgType,
                                     eLoggerMsgType.Separator,
                                     p_Message);

              //========================================================================
              #endregion
              break;

            case eLoggerMsgType.StopSeparator :
              #region StopSeparator
              //========================================================================

              WriteHTMLSeparatorLine(sw,
                                     p_MsgID,
                                     p_MsgType,
                                     eLoggerMsgType.Head,  // MsgStyle!
                                     p_Message);

              sw.Write(HTML.TableEnd());
              sw.Write(HTML.BR());

              sw.Write(Environment.NewLine);

              //========================================================================
              #endregion
              break;

            default:
              #region default
              sw.Write(HTML.TRStart());

              //=== ID =================================================================

              string sID = "";

              if(p_MsgID <= 0)
              {
                sw.Write(HTML.TD(sID, p_MsgType.ToString()));
              }
              else
              {
                sw.Write(HTML.TD(p_MsgID.ToString("000000#"), p_MsgType.ToString()));
              }

              //=== Date ===============================================================

              sw.Write(HTML.TD(LogTimeToString(DateTime.Now), p_MsgType.ToString())); //,200

              //=== MSG type ===========================================================

              #region c
  //            if(p_MsgType == eLoggerMsgType.none)
  //            {
  //              sw.Write(HTML.TD("", p_MsgType.ToString()));
  //            }
  //            else
  //            {
  //              sw.Write(HTML.TD(p_MsgType.ToString(), p_MsgType.ToString()));
  //            }
  #endregion

              //=== Message ============================================================

              string sFNIN    = "==>";
              string sFNOUT   = "<==";
              string sMsgType = "";
              string postfix  = "";

              if(p_MsgType == eLoggerMsgType.none)
              {
                // add Msg type + message to the log
                sw.Write(HTML.TD("",        p_MsgType.ToString()));
                sw.Write(HTML.TD(p_Message, p_MsgType.ToString() + "T"));
              }
              else
              {
                switch(p_MsgType)
                {
                  case eLoggerMsgType.FNTraceIN :
                    sMsgType = sFNIN;
                    postfix  = "A";
                    break;

                  case eLoggerMsgType.FNTraceOUT :
                    sMsgType = sFNOUT;
                    postfix  = "A";
                    break;

                  default :
                    sMsgType = p_MsgType.ToString();
                    break;
                }

                // add Msg type + message to the log
                sw.Write(HTML.TD(sMsgType,              p_MsgType.ToString() + postfix));
                sw.Write(HTML.TD("[" + p_Message + "]", p_MsgType.ToString() + "T"));
              }

              //========================================================================

              sw.Write(HTML.TREnd());

  //            sw.Write(Environment.NewLine);

              //========================================================================
              #endregion
              break;
          }
          ok = true;
        }
        catch
        {
          ok = false;
        }
        finally
        {
          sw.Close();
        }
        #endregion
      }
      //========================================================================

      return ok;
    }
    #endregion

    //----------------------------------------------------------------------

    #endregion

  //---------------------------------------------------------------------------
  }
}
