using System;
using System.Web;
using System.Net;
using System.IO;
using System.Collections;
using Microsoft.Win32;
using System.Reflection;
using System.Diagnostics;
using System.Threading;
using System.Windows.Forms;

using JWorldBuilder.Logging;
using FTPLib;

namespace JadamPrograms.TeamServer
{
  public delegate void ProgressChanged(int  CurProgress,
                                       long FileSize,
                                       long TotalBytesLeft);
                                       
  public delegate void FilesChanged(ArrayList Files);
  public delegate void NewMessageDownloaded(string Message);
   
                                       
  public enum Level : int
  {
    SimpleUser = 4,
    Developer = 3,
    ProjectMaster = 2,
    Administrator = 1,
    nothing = 0
  }
                                        
	/// <summary>
	/// Web Handler
	/// Team Server-hez - CheckOut(); CheckIn(); GetFileNamesFromProjServer();
	/// </summary>
	public class WebHandler
  {
    #region Start
    #region Vltozk
    public bool LogedIn = false;
    
    public string ok = "ok ret";
    
    //
    //Ha egy fgvny visszatrsi rtke false akkor ebben a vltozban tallhat a hibazenet.
    //
    public string LastError = "";
 	  
    //
    //Az FTP Server hasznlathoz szksges osztly.
    //
    public FTP ftp = new FTP();
	  
    //
    //Fjl feltlts s letltskor(Ki-,Be- csekkols) a progressbar hasznlathoz szksges adatokat szolgltat 
    //esemny.
    //
    public event ProgressChanged ProgressChangedEvent;
    
    public event FilesChanged FilesChangedEvent;    
	  public event NewMessageDownloaded NewMessageEvent;
	      
    //
    //A log rshoz szksges osztly.
    //
    Global g = Global.Instance;
    public Logger log = new Logger(Application.StartupPath + "\\WebHandler_Log.html",
                                   Application.StartupPath + "\\LoggerStyle.css");
	  
    //
    //A Team Server homelap cme.
    //
    public string ServerPath = "";

    //
    //FTP-hez szksges adatok.
    //
    public string server = "";
    public string user   = "";
    public string pass   = "";
    public int    port   = 21;
    
    //
    //Login utn kapott szm mely azonostja a felhasznlt.
    //
    public int UserId = 0;
    public string LoginedUserName = "";
    
    //
    //Belltsok
    //
    public int TimeToWait = 10000; 
    
    //
    //A serveren tallhat fileok knyvtra.
    //
    public string FtpPath = "wwwroot/Files";
    
    public ArrayList FileNames = new ArrayList(); 
    
    public RichTextBox rtb;    
    #endregion
    #region Constructor
    public WebHandler(string pServerPath,
      string pServer,
      string pUser,
      string pPass,
      int    pPort)
    {
      ServerPath = pServerPath;
      server     = pServer;
      user       = pUser;
      pass       = pPass;
      port       = pPort;
      log.OpenLog();
    }
    #endregion      
    #endregion
		
    #region Homepage
    #region GetDataFromServer
    
    //
    //PHP-vel adotak kr a szervertol.
    //
    public string GetDataFromServer(string QueryString)
    {
      try
      {
        //MessageBox.Show(QueryString,"Information");
        WebRequest reqest = WebRequest.Create(QueryString);
        WebResponse response = reqest.GetResponse();
        StreamReader sr = new StreamReader(response.GetResponseStream(),System.Text.Encoding.UTF8);
        string read = sr.ReadToEnd();
        reqest.Abort();
        response.Close();
        sr.Close();
        return read;
      }
      catch(System.Exception e)
      {
        LastError = e.Message;
        return "";
      }
    }
    #endregion
    #region PHPCommands
    #region PHPLogin
    public string PHPLogin(string user,
                           string pass)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=Login&User=" + user + "&Pass=" + pass);
    }
    #endregion
    #region PHPLogout
    public string PHPLogout()
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=Logout&UserId=" + UserId);
    }
    #endregion   
    #region PHPCheckIn
    public string PHPCheckIn(string FileName)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=CheckIn&FileName=" + FileName + "&UserId=" + UserId);
    }
    #endregion   
    #region PHPCheckOut
    public string PHPCheckOut(string FileName)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=CheckOut&FileName=" + FileName + "&UserId=" + UserId);
    }
    #endregion     
    #region PHPGetFiles
    public string PHPGetCheckedOutFiles()
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=GetCheckedOutFiles&UserId=" + UserId);
    }
    #endregion        
    #region PHPGetMessages
    public string PHPGetMessages(bool OnlyNew)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=GetMessages&OnlyNew=" + OnlyNew + "&UserId=" + UserId);
    }
    #endregion 
    #region PHPWriteMessage
    public string PHPWriteMessage(string ToUser,
                                  string Message)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=NewMessage&To=" + ToUser + "&Msg=" + Message + "&UserId=" + UserId);
    }
    #endregion         
    #region PHPGetUsers
    public string PHPGetUsers()
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=GetUsers&UserId=" + UserId);
    }
    #endregion            
    #region PHPAddUser
    public string PHPAddUser(string name,
                             string pass,
                             string email,
                             Level  level)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=NewUser&Name=" + name + "&Pass=" + pass + "&EMail=" + email + "&Level=" + (int)level + "&UserId=" + UserId);
    }
    #endregion             
    #region PHPDelUser
    public string PHPDelUser(string id)
    {
      return GetDataFromServer(ServerPath + "/teamserv.php?Comm=DelUser&Id=" + id + "&UserId=" + UserId);
    }
    #endregion               
    #endregion
    #region PHPValidate
    public bool PHPValidate(string Ret)
    {
      if(Ret.IndexOf(ok) != -1)
      {
        return true;
      }
      else
      {
        LastError = GetErrorMsg(Ret);
        return false;
      }
    }    
    #endregion
    #region GetErrorMsg
    public string GetErrorMsg(string text)
    {
      string ErrCode = text.Substring(text.IndexOf("err[")+4,6);    
      return ErrCode + ": " + PHPErrCodeToMsg(ErrCode);
    }    
    #endregion
    #region PHPErrCodeToMsg
    public string PHPErrCodeToMsg(string ErrorCode)
    {
      switch(ErrorCode)
      {
        case "TSE000" :
          return "Kapcsolds az adatbzishoz sikertelen!";
          break;
        case "TSE001" :
          return "Adatbzis vlaszts sikertelen!";
          break;          
        case "TSE002" :
          return "Az adatok trolsa sikertelen!";
          break;
        case "TSE003" :
          return "A megadott nv vagy jelsz hibs.Hozzfrs megtagadva!";
          break;
        case "TSE004" :
          return "A kijeletkezs sikertelen!";
          break;          
        case "TSE005" :
          return "A becsekkols sikertelen!";
          break;          
        case "TSE006" :
          return "A vlasztott filet mr kicsekkoltk.Krem vrjon amg becsekkoljk!";
          break;         
        case "TSE007" :
          return "Kicsekkols sikertelen.";
          break;          
        case "TSE008" :
          return "A felhasznl nem azonosthat: Nincs felhasznl azonost.";
          break;
        case "TSE009" :
          return "A felhasznlnv, vagy a jelsz hinyzik!";
          break;          
        case "TSE010" :
          return "A filenv hinyzik!";
          break;          
        case "TSE011" :
          return "Az OnlyNew hinyzik!";
          break;         
        case "TSE012" :
          return "A To, vagy az zenet hinyzik!";
          break;           
        case "TSE013" :
          return "A felhasznl adatai hinyoznak!";
          break;
        case "TSE014" :
          return "A felhasznl neve hinyzik!";
          break;          
        case "TSE015" :
          return "j felhasznl ltrehozs a sikertelen!";
          break;          
        case "TSE016" :
          return "A UserId hibs.Hozzfrs megtagadva!";
          break;         
        case "TSE017" :
          return "Nincs megfelelo jogosultsga a vgrehajtshoz!";
          break;           
        default :
          return "Ismeretlen hibakd. Hibakd:" + ErrorCode; 
          break;   
      }
    }    
    #endregion
    #endregion
		
    #region Login & Logout
    #region Login
    public bool Login(string User,
                      string Pass)
    {
      if(!LogedIn)
      {
        string Ret = PHPLogin(User,Pass);  
        if(Ret.IndexOf(ok) != -1)
        {
          try
          {
            UserId = Convert.ToInt32(Ret.Substring(Ret.IndexOf(ok)+7));
            LoginedUserName = User;
            FileNames = new ArrayList();
            InfoDownloaderThread = new Thread(new ThreadStart(InfoDownloader));
            InfoDownloaderThread.Name = "JAP Team Server - InfoDownloaderThread";
            InfoDownloaderThread.Start();
            LogedIn = true; 
            return true;
          }
          catch(System.Exception e)
          {
            UserId = 0;
            LoginedUserName = "";
            LastError = "Konvertlsi hiba.";
            return false;
          }
        }
        else
        {
          LastError = GetErrorMsg(Ret);
          return false;
        }
      }
      else
      {
        LastError = "Krem jelentkezzen ki mieltt jra bejelentkezne!";
        return false;
      }
    }
    #endregion  
    #region Logout
    public bool Logout()
    {
      if(LogedIn)
      {
        try
        {   
          InfoDownloaderThread.Suspend();
        }
        catch
        {}
        string Ret = PHPLogout(); 
        if(Ret.IndexOf(ok) != -1)
        {
          try
          {
            UserId = 0;
            LoginedUserName = "";
            InfoDownloaderThread.Abort();
            FileNames = new ArrayList();
            return true;
          }
          catch(System.Exception e)
          {
            UserId = 0;
            LoginedUserName = "";
            return false;
          }
        }
        else
        {
          LastError = GetErrorMsg(Ret);
          try
          {
            InfoDownloaderThread.Resume();
          }
          catch
          {}
          return false;
        }
      }
      else
      {
        LastError = "Nincs bejelentkezett felhasznl!";
        return false;
      }
    }
    #endregion  
 		#endregion	
		
    #region FTP
    #region CheckOut
    public bool CheckOut(string  FileName,
		                     string  SavePath)
		{
      string Ret = PHPCheckOut(FileName);
      if(Ret.IndexOf(ok) != -1)
      {
        try
        {
          
          ftp.server = server;
          ftp.user   = user;
          ftp.pass   = pass;
          ftp.port   = 21;    
          ftp.Connect();
          
          ftp.ChangeDir("/wwwroot");
          ftp.ChangeDir("/wwwroot/Files");  
          
          ftp.OpenDownload(FileName,SavePath,true);   
          
          try
          {
            g.CheckedOutFiles.Add(FileName); 
          }
          catch
          {}  
          return true;
        }
        catch(System.Exception e)
        {
          LastError = e.Message;
          return false;
        }
      }
      else
      {
        LastError = GetErrorMsg(Ret);
        return false;
      }
		}
		
    public bool CheckOutNoEvent(string FileName,
                                string SavePath)
    {
      string Ret = PHPCheckOut(FileName);
      if(Ret.IndexOf(ok) != -1)
      {
        try
        {
          
          ftp.server = server;
          ftp.user   = user;
          ftp.pass   = pass;
          ftp.port   = 21;    
          ftp.Connect();
          
          ftp.ChangeDir("/wwwroot");
          ftp.ChangeDir("/wwwroot/Files");  
          
          ftp.OpenDownload(FileName,SavePath,true);   
          
          try
          {
            g.CheckedOutFiles.Add(FileName); 
          }
          catch
          {}  
          return true;
        }
        catch(System.Exception e)
        {
          LastError = e.Message;
          return false;
        }
      }
      else
      {
        LastError = GetErrorMsg(Ret);
        //LastError = Ret;
        return false;
      }
    }
    
    public bool DoCheckOut(ref int  CurProgress,
                           ref long TotalLeftBytes,
                           ref long FileSize)
    { 
      int perc = 0;
			
      while(ftp.DoDownload() > 0)
      {
        perc = (int)(((ftp.BytesTotal) * 100) / ftp.FileSize);
        CurProgress = perc;
        FileSize = ftp.FileSize;
        TotalLeftBytes = ftp.BytesTotal;
        return true;
      }
          
      ftp.Disconnect();   
          
      
      return false;  
    }
		#endregion		
    #region CheckIn
    public bool CheckIn(string FileName,
                        string FileNameOnServer)
    {
      string Ret = PHPCheckIn(FileNameOnServer);
      if(Ret.IndexOf(ok) != -1)
      {
        try
        {
          
          ftp.server = server;
          ftp.user   = user;
          ftp.pass   = pass;
          ftp.port   = 21;         
          ftp.Connect();
          
          ftp.ChangeDir("/wwwroot");
          ftp.ChangeDir("/wwwroot/Files");  
          
          ftp.OpenUpload(FileName,FileNameOnServer,false);
          int perc = 0;
			
          while(ftp.DoUpload() > 0)
          {
            perc = (int)(((ftp.BytesTotal) * 100) / ftp.FileSize);
            ProgressChangedEvent(perc, ftp.FileSize, ftp.BytesTotal);
          }  
          
          ftp.Disconnect();
          
          try
          {
            g.CheckedOutFiles.Remove(FileName); 
          }
          catch
          {}             
          return true;
        }
        catch(System.Exception e)
        {
          //GetDataFromServer("http://jap-projserv.extra.hu/teamserv.php?Check=Out&FileName=" + FileNameOnServer);
          LastError = e.Message;
          return false;
        }
      }
      else
      {
        LastError = GetErrorMsg(Ret);
        return false;
      }
    }
    
    public bool CheckInNoEvent(string FileName,
                               string FileNameOnServer)
    {
      string Ret = PHPCheckIn(FileNameOnServer);
      if(Ret.IndexOf(ok) != -1)
      {
        try
        {
          
          ftp.server = server;
          ftp.user   = user;
          ftp.pass   = pass;
          ftp.port   = 21;         
          ftp.Connect();
          
          ftp.ChangeDir("/wwwroot");
          ftp.ChangeDir("/wwwroot/Files");  
          
          ftp.OpenUpload(FileName,FileNameOnServer,false);
          try
          {
            g.CheckedOutFiles.Remove(FileName); 
          }
          catch
          {}  
          return true;
        }
        catch(System.Exception e)
        {
          LastError = e.Message;
          return false;
        }
      }
      else
      {
        LastError = GetErrorMsg(Ret);
        return false;
      }
    }
    
    public bool DoCheckIn(ref int CurProgress,
                          ref long TotalLeftBytes,
                          ref long FileSize)
    {             
      int perc = 0;
			
      while(ftp.DoUpload() > 0)
      {
        perc = (int)(((ftp.BytesTotal) * 100) / ftp.FileSize);
        CurProgress = perc;
        FileSize = ftp.FileSize;
        TotalLeftBytes = ftp.BytesTotal;
        return true;
      }  
          
      ftp.Disconnect();
           
      return false;  
    }
    #endregion
    #region GetFileNamesFromTheProjServer
    public bool GetFileNamesFromTheProjServer(ref ArrayList FileNames)
    {
      try
      {
        ftp.server = server;
        ftp.user   = user;
        ftp.pass   = pass;
        ftp.port   = 21;         
        ftp.Connect();
          
        ftp.ChangeDir("/wwwroot");
        ftp.ChangeDir("/wwwroot/Files");  
          
        ArrayList FNames = ftp.ListFiles();
        FileNames = (ArrayList)FNames.Clone();
        string[] Names = PHPGetCheckedOutFiles().Split('|'); 
        int i = 0;
        foreach(string s in FNames)
        {
          string CheckState = "Checked In";
          string Name = s.Substring(s.LastIndexOf(" ")+1);
          foreach(string n in Names)
          {
            string[] ns = n.Split('#');
            if(Name == ns[0])
            {
              CheckState = "Checked Out(" + ns[1] + ")";
              break;
            }
          }
          FileNames[i] = Name + "|" + CheckState;
          i++;
        }
        
        ftp.Disconnect();
        return true;
      }
      catch(System.Exception e)
      {
        LastError = e.Message;
        return false;
      }
    }
    #endregion
    #endregion
    
    #region Thread to refresh
    public int LogLevel = 5;
    public void WriteLine(string text,
                          int level)
    {
      if(level >= LogLevel)
      {
        if(rtb != null)
        {
          rtb.Text += text + Environment.NewLine;
          rtb.Refresh();
        }
        log.WriteInfo("InfoDownloaderThread => " + text);
      }
    }
    
    public void InfoDownloader()
    {
      try
      {
        while(true)
        {
          try
          {
            WriteLine("Frissts megkezdse...",5);
            FTP ftpt = new FTP(server,
                               port,
                               user,
                               pass);
             
            WriteLine("Kapcsolds a serverhez...",4);      
            ftpt.Connect();
            
            WriteLine("Knyvtrvlts...",4);
            ftpt.ChangeDir(FtpPath);
            
            WriteLine("Filenevek letltse...",5);
            ArrayList getted_FNames = ftpt.ListFiles();
            ArrayList FNames = (ArrayList)getted_FNames.Clone();
            
            WriteLine("Kapcsolat megszntetse...",4);   
            ftpt.Disconnect();
            
            WriteLine("Kicsekkolt filenevek letltse...",3);
            string RetOFNames = PHPGetCheckedOutFiles();
            WriteLine("RetOFnames = " + RetOFNames,1);
            WriteLine("RetOFnames Mdostott = " + RetOFNames.Substring(RetOFNames.IndexOf(ok)+6),1);     
            string[] Names = RetOFNames.Substring(RetOFNames.IndexOf(ok)+7).Split('|');
            WriteLine("Kicsekkolt fileok szma: " + Names.Length,3);
            
            WriteLine("Keress a filenevekben...",3);
            int i = 0;
            string[] ns;
            foreach(string s in getted_FNames)
            {
              string CheckState = "Checked In";
              string Name = s.Substring(s.LastIndexOf(" ")+1);
              foreach(string n in Names)
              {
                ns = n.Split('#');
                WriteLine("ns sztvlasts eredmnye: 0:" + ns[0] + " 1:" + ns[1],1);
                if(Name == ns[0])
                {
                  CheckState = "Checked Out(" + ns[1] + ")";
                  if(LoginedUserName == ns[1])
                  {
                    g.CheckedOutFiles.Add(ns[0]);
                  }
                  break;
                }
              }
              FNames[i] = Name + "|" + CheckState;
              
              WriteLine(FNames[i].ToString(),2);
                      
              i++;
            } 
            
            WriteLine("szzehasonlts...",4);
            WriteLine("Fileok szma: " + FileNames.Count,3);          
            bool Modifyed = false;
            if(FNames.Count == FileNames.Count)
            {
              for(int i2 = 0;i2 < FileNames.Count;i2++)
              {
                if(FNames[i2] != null)
                {
                  WriteLine(i2 + ": " + FNames[i2] + "     ",1);   
                  if(FileNames[i2] != null)
                  {
                    WriteLine(i2 + ": " + FNames[i2],1);
                    if(FNames[i2].ToString() != FileNames[i2].ToString())
                    {
                      Modifyed = true;
                      break;
                    }
                    else
                    {
                      WriteLine("Nem mdosult",3);   
                      continue;
                    }
                  }
                  else
                  {
                    Modifyed = true;
                    break;
                  }
                }
                else
                {
                  Modifyed = true;
                  break;
                }
              }
            }
            else
            {
              Modifyed = true;
            }
            if(Modifyed)
            {
              WriteLine("Mdosult!!!",5);
              FileNames = FNames;
              FilesChangedEvent(FNames);
            }
          }
          catch(System.Exception e)
          {
            WriteLine("Hiba:" + e.Message,5);   
          }
          
//          string Ret = PHPGetMessages(true);
//          if(Ret.IndexOf(ok) != -1)
//          {
//            string Message = Ret.Substring(Ret.IndexOf(ok)+7);
//            if(Message != "")
//            {
//              NewMessageEvent(Message);
//            }
//          }
          
          WriteLine("Kvetkezo frists: " + TimeToWait + " tizedmsodperc mlva.",5);
          Thread.Sleep(TimeToWait);                   
        }
      }
      catch(System.Exception e)
      {
        WriteLine(e.Message,5);
        System.Windows.Forms.MessageBox.Show(e.Message,"Info Downloader Thread auto-aborted");
        InfoDownloaderThread.Abort();
      }
    }
    Thread InfoDownloaderThread;
    #endregion
  }
}
