/*
	test file

*/

//#include <afxdao.h>
#include <windows.h>
#include <stdio.h>

#include <stdio.h>
#include <conio.h>
#include <limits.h>
#include <math.h>



#include "..\engine\3dstate.h"

#define PLANE_MAX_SPEED 1000
#define PLANE_MIN_SPEED 100
//#define PLANE_DROP_DOWN_SPEED 200 //The speed in which the plane is starting to lose altitude
#define PLANE_MIN_HEIGHT 6000
#define CAMERA_MIN_HEIGHT 2000

#define DEFAULT_FOG_VISIBILITY 2500000

void turn_plane_according_to_keyboard_input(DWORD plane_object);
double change_speed_according_to_keyboard_input(double speed, DWORD font_handle);
int make_sure_the_plane_doesnt_go_below_the_ground(DWORD plane_object, DWORD font_handle);
void change_shadow_settings_according_to_keyboard_input(DWORD plane_object, DWORD ariel18);
void change_sound_settings_according_to_keyboard_input(DWORD background_music, DWORD propellor_sound, DWORD ariel18);
void change_camera_properties_according_to_keyboard_input(DWORD camera);
void change_fog_settings_according_to_keyboard_input(void);

#pragma warning(disable : 4100)
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR     lpCmdLine,  int nCmdShow)
{
	
	//For debugging purposes. Should be removed in release version
	//STATE_engine_write_error_messages_to_the_screen(ON);

	//Set Background music
	DWORD background_music=STATE_sound_load("..\\sound\\Elgar.mp3", NULL,SOUND_DISTANCE_DEFAULT);
	STATE_sound_play(background_music, SOUND_LOOP_NORMAL);
	DWORD ariel18=STATE_font_create("Ariel", 18, FONT_CREATE_DEFAULT);
	STATE_font_write_text(ariel18, "Background music ON", 0, 40, 255, 50, 255, 255, FONT_WRITE_DEFAULT, INT_MAX);


	//Load the world
	int rc=STATE_engine_load_world("..\\models\\dothan.3dstate","", "..\\models\\bitmaps", USER_DEFINED_BEHAVIOR);
	if(rc==VR_ERROR) {
		printf("\nFailed to load world.\n");
		printf("Please look at error.log to see why.\n");
		printf("Press any key to exit\n");
		getch();
		return(-1);
	}

	//STATE_sound_stop(background_music);

	//load the airoplane
	DWORD plane_object=STATE_object_create_from_file("..\\models\\airplane 2.x");
	if(plane_object==NULL)
		return(-1); //safty

	double scale[3]={1000, 1000, 1000}; //make the plane bigger
	STATE_object_set_scale(plane_object, scale);

	//The object is turned around so lets turn it 180 degrees
	//This is important so that the STATE_chase functions work as
	//we expect them to work
	STATE_object_rotate_z(plane_object, 180, WORLD_SPACE);
	STATE_object_redefine_axis_system(plane_object);

	//Get the world dimesions so that we can watch later
	//that the plane doesn't get out of the world ...
	double world_boundaries[2][3];
	STATE_group_get_bounding_box(NULL, world_boundaries);

	STATE_font_write_text(ariel18, "Keys: F1,  Arrow keys,  1,2,3,4,5,   Q,W,S,   A,S,  D,F,  G,H,  J,K   B, Space bar", 0, 20, 255, 0, 0, 140, FONT_WRITE_DEFAULT, INT_MAX);

	//Just one line is needed to add the shadows to the plane !!!
	STATE_object_auto_add_shadows(plane_object, ON, 1);
	STATE_font_write_text(ariel18, "Shadow ON", 0, 35, 255, 255, 50, 255, FONT_WRITE_DEFAULT, INT_MAX);

	//Set the sound of the wind/plane
	DWORD propellor_sound=STATE_sound_load("..\\sound\\propeller_sound.wav", plane_object, 600000);
	STATE_sound_set_volume(propellor_sound, 800);
	STATE_sound_play(propellor_sound, SOUND_LOOP_NORMAL);
	STATE_font_write_text(ariel18, "Wind noise: ON", 0, 45, 255, 0, 255, 0, FONT_WRITE_DEFAULT, INT_MAX);


	//Put some ground details texture
	DWORD ground_bmp=STATE_bitmap_load("..\\models\\Bitmaps\\second_bitmap.bmp", -1);
	if(	ground_bmp!=NULL) {
		//Try cancel this line or putting a different number than 100 and see what it does
		STATE_group_wrap_a_bitmap(NULL, ground_bmp, 35, 35, 1);
	}


	//Add a light source
	double location[3]={0,0, 10000}; //put it up in the sky like the sun
	DWORD m_light=STATE_light_create("light1", location);
	STATE_light_set_type_of_light(m_light, LIGHT_AMBIENT | LIGHT_DIFFUSE);
	//STATE_light_set_ray_tracing(m_light, YES);
	STATE_light_set_ambient(m_light, 0.8);
	STATE_light_set_specular(m_light, 0);
	STATE_light_set_diffuse(m_light, 3);
//	double direction[3]={0,0, -1};
//	STATE_light_set_direction(m_light, direction);
	STATE_light_activate(m_light, LIGHT_SET);

	location[2]=10000000; //this height will look better with the plane shadows since is simulate the sun and the sun is very high above ...
	STATE_light_set_location(m_light, location);

	
	STATE_engine_maximize_default_rendering_window();
	
	STATE_engine_show_frames_per_second_rate(INFORMATION_ALL);


	//Set the camera (point of view) 10000 VR units above the ground
	DWORD camera=STATE_camera_get_default_camera();

	
	//Setting the camera to follow the plane automatically
	STATE_camera_set_chase_type(camera, CHASE_FLEXIBLE);
	//STATE_camera_set_chase_type(camera, CHASE_PRECISE);
	STATE_camera_set_object_to_chase(camera, plane_object);
	STATE_camera_set_chase_distance(camera, 40000); //set the camera behind the airplane
	double offset[3]={0, 0, 10000}; //set an offset so it is always a bit above
	STATE_camera_set_chase_offset(camera, offset);

	STATE_engine_set_fog(ON,  220, 220, 180, FOG_MODE_AUTO, DEFAULT_FOG_VISIBILITY, FOG_START_DISTANCE_DEFAULT, 0);

	double speed=PLANE_MAX_SPEED/3;

	MessageBox(NULL, "Instructions:\n\nPress F1 To get this help popup\n\nUse the arrow keys to navigate the plane\n\nUse Q,W,E to change the camera modes\n\nUse 1,2,3,4,5 to control the camera distance\n\nUse A,S to toggle shadow modes\n\nUse D,F,G,H to toggle sound modes\n\nUse J,K to control fog intensity\n\nUse the space bar and 'B' to increase or decrease the speed\n\nThings to notice\n------------------\n1. Press E (NO_CHASE) and notice how the the sound gets automatically stronger when the plane gets closer to you\nNotice that when the plane is on the right, you hear it more from the right speaker\n(make sure your speaker are positioned correctly)\n\n2. Press W and then 2 and notice the shadows on the plane wing and tail when turning\n", "Instructions", MB_ICONINFORMATION);


	while( (GetAsyncKeyState(VK_ESCAPE)&1) ==0 ) {

		STATE_engine_render(NULL,camera);

		STATE_object_move(plane_object, OBJECT_SPACE, -speed, 0,0); //move the plane forwad. This is towards the plane negative X axis

		turn_plane_according_to_keyboard_input(plane_object);

		speed=change_speed_according_to_keyboard_input(speed, ariel18);


		//make sure the plane doesnt go out of the world
		STATE_object_get_location1(plane_object, location);
		for(int i=0; i<2; i++) { //note that <2 and not <3 because we don't care about getting out of the world on the X axis
			if( (location[i]>world_boundaries[1][i]*0.9) || (location[i]<world_boundaries[0][i]*0.9)) {
				STATE_object_set_location(plane_object, 0,0, 5000);
			}
		}		

		make_sure_the_plane_doesnt_go_below_the_ground(plane_object, ariel18);
		
		//make sure that the camera doesn't get into the ground
		double height;
		DWORD ground_poly=STATE_camera_get_height_above_ground(camera, 0, &height);
		if(ground_poly!=NULL) {
			if(height<CAMERA_MIN_HEIGHT) {
				STATE_camera_move(camera, WORLD_SPACE, 0,0, CAMERA_MIN_HEIGHT-height);
			}
		}

		change_camera_properties_according_to_keyboard_input(camera);

		change_shadow_settings_according_to_keyboard_input(plane_object, ariel18);
		
		change_sound_settings_according_to_keyboard_input(background_music, propellor_sound, ariel18);

		change_fog_settings_according_to_keyboard_input();

		//Show the help popup
		if(GetAsyncKeyState(VK_F1)<0) {
			MessageBox(NULL, "Instructions:\n\nPress F1 To get this help popup\n\nUse the arrow keys to navigate the plane\n\nUse Q,W,E to change the camera modes\n\nUse 1,2,3,4,5 to control the camera distance\n\nUse A,S to toggle shadow modes\n\nUse D,F,G,H to toggle sound modes\n\nUse the space bar and 'B' to increase or decrease the speed\n\nThings to notice\n------------------\n1. Press E (NO_CHASE) and notice how the the sound gets automatically stronger when the plane gets closer to you\nNotice that when the plane is on the right, you hear it more from the right speaker\n(make sure your speaker are positioned correctly)\n\n2. Press W and then 2 and notice the shadows on the plane wing and tail when turning\n", "Instructions", MB_ICONINFORMATION);
		}


	}

	STATE_engine_close();

	return(0);
}
#pragma warning(default : 4100)

//note that the code here take care that 
//when keyboard keys are not pressed the plane would get straight automatically
void turn_plane_according_to_keyboard_input(DWORD plane_object)
{
	//note tat rotation around our airplane Y axis increases the tilt angle (how much the nose is higher or lower than the tail)
	if(GetAsyncKeyState(VK_DOWN)<0)	STATE_object_rotate_y(plane_object, 1,OBJECT_SPACE);
	if(GetAsyncKeyState(VK_UP)<0)	STATE_object_rotate_y(plane_object, -1,OBJECT_SPACE);

	//note tat rotation around the airplane X axis increases the bank angle (when one wing is higher and the other wind is lower)
	if(GetAsyncKeyState(VK_LEFT)<0) STATE_object_rotate_x(plane_object, 1,OBJECT_SPACE);
	if(GetAsyncKeyState(VK_RIGHT)<0) STATE_object_rotate_x(plane_object, -1,OBJECT_SPACE);

	double bank=STATE_object_get_bank(plane_object);

	//Make the turn proportional to the bank angle. The more one wing is higher and the other one is lower the turn
	//will be bigger
	STATE_object_rotate_z(plane_object, -bank*0.02,OBJECT_SPACE);
	
	//decrease the bank == try to straighten the plane
	bank*=0.98;
	STATE_object_set_bank(plane_object, bank);

	//try to reduce the tilt and thus straighten the plane (so that the nose is at the same hight as the tail)
	double tilt=STATE_object_get_tilt(plane_object);
	tilt*=0.9;
	STATE_object_set_tilt(plane_object, tilt);
}




double change_speed_according_to_keyboard_input(double speed, DWORD font_handle)
{
	if(GetAsyncKeyState(' ')<0)
		speed+=1;

	if(GetAsyncKeyState('B')<0)
		speed-=1;

	if(speed>PLANE_MAX_SPEED)
		speed=PLANE_MAX_SPEED;

	if(speed<PLANE_MIN_SPEED)
		speed=PLANE_MIN_SPEED;

	//Write speed to screen
	char speed_str[1000];
	sprintf(speed_str, "Speed: %ld", (int)speed);
	STATE_font_write_text(font_handle, speed_str, 0, 25, 255, 255, 255, 120, FONT_WRITE_DEFAULT, INT_MAX);


	return(speed);
}


int make_sure_the_plane_doesnt_go_below_the_ground(DWORD plane_object, DWORD font_handle)
{
	 static int number_of_possible_crashes=0;	
	
	//Make sure the plane doesn't get into the ground
	double height=STATE_object_get_height(plane_object, YES);
	if(height<PLANE_MIN_HEIGHT) {
		STATE_object_move(plane_object, WORLD_SPACE, 0,0, PLANE_MIN_HEIGHT-height);
		number_of_possible_crashes++;
		//speed*=0.98; //slow down

		char crashes_text[1000];
		sprintf(crashes_text, "Possible Crashes: %ld", (int)number_of_possible_crashes);
		STATE_font_write_text(font_handle, crashes_text, 0, 30, 255, 255, 255, 120, FONT_WRITE_DEFAULT, INT_MAX);
	}

	return(number_of_possible_crashes);
}


void change_camera_properties_according_to_keyboard_input(DWORD camera)
{
	if(GetAsyncKeyState('1')<0)
		STATE_camera_set_chase_distance(camera, 15000); //set the camera behind the airplane

	if(GetAsyncKeyState('2')<0)
		STATE_camera_set_chase_distance(camera, 30000); //set the camera behind the airplane

	if(GetAsyncKeyState('3')<0)
		STATE_camera_set_chase_distance(camera, 45000); //set the camera behind the airplane

	if(GetAsyncKeyState('4')<0)
		STATE_camera_set_chase_distance(camera, 60000); //set the camera behind the airplane

	if(GetAsyncKeyState('5')<0)
		STATE_camera_set_chase_distance(camera, 75000); //set the camera behind the airplane


	if(GetAsyncKeyState('Q')<0)
		STATE_camera_set_chase_type(camera, CHASE_FLEXIBLE);

	if(GetAsyncKeyState('W')<0)
		STATE_camera_set_chase_type(camera, CHASE_PRECISE);

	if(GetAsyncKeyState('E')<0)
		STATE_camera_set_chase_type(camera, NO_CHASE);

}

void change_shadow_settings_according_to_keyboard_input(DWORD plane_object, DWORD ariel18)
{

	if(GetAsyncKeyState('A')<0) {
		STATE_object_auto_add_shadows(plane_object, ON, 1);
		STATE_font_write_text(ariel18, "Shadow ON", 0, 35, 255, 255, 50, 255, FONT_WRITE_DEFAULT, INT_MAX);
	}

	if(GetAsyncKeyState('S')<0) {
		STATE_object_auto_add_shadows(plane_object, OFF, 1);
		STATE_font_write_text(ariel18, "Shadow OFF", 0, 35, 255, 255, 50, 255, FONT_WRITE_DEFAULT, INT_MAX);
	}
}


void change_sound_settings_according_to_keyboard_input(DWORD background_music, DWORD propellor_sound, DWORD ariel18)
{
	if(GetAsyncKeyState('D')<0) {
		if(STATE_sound_is_sound_playing(background_music)==NO)
			STATE_sound_play(background_music, SOUND_LOOP_NORMAL);
		STATE_font_write_text(ariel18, "Background music ON", 0, 40, 255, 50, 255, 255, FONT_WRITE_DEFAULT, INT_MAX);
	}

	if(GetAsyncKeyState('F')<0) {
		if(STATE_sound_is_sound_playing(background_music)==YES)
			STATE_sound_stop(background_music);
		
		STATE_font_write_text(ariel18, "Background music OFF", 0, 40, 255, 50, 255, 255, FONT_WRITE_DEFAULT, INT_MAX);
	}


	if(GetAsyncKeyState('G')<0) {
		if(STATE_sound_is_sound_playing(propellor_sound)==NO)
			STATE_sound_play(propellor_sound, SOUND_LOOP_NORMAL);
		STATE_font_write_text(ariel18, "Wind noise: ON", 0, 45, 255, 0, 255, 0, FONT_WRITE_DEFAULT, INT_MAX);
	}

	if(GetAsyncKeyState('H')<0) {
		if(STATE_sound_is_sound_playing(propellor_sound)==YES)
			STATE_sound_stop(propellor_sound);
		
		STATE_font_write_text(ariel18, "Wind noise: OFF", 0, 45, 255, 0, 255, 0, FONT_WRITE_DEFAULT, INT_MAX);
	}
}

#define MAX_FOG 1e8
#define MIN_FOG 2e5

//decrease or increase the fog intensity according to the keyboard input
void change_fog_settings_according_to_keyboard_input()
{
	static double fog_visibility=DEFAULT_FOG_VISIBILITY;
	
	double factor=1;
	if(GetAsyncKeyState('J')<0) {
		factor=0.95;
	}
	if(GetAsyncKeyState('K')<0) {
		factor=1.05;
	}

	if(factor==1)
		return; //no change is needed. Neither of the keys is pressed

	fog_visibility*=factor;

	
	if(fog_visibility<MIN_FOG)
		fog_visibility=MIN_FOG;

	DWORD ariel18=STATE_font_create("Ariel", 18, FONT_CREATE_DEFAULT);
	
	if(fog_visibility>MAX_FOG) {
		fog_visibility=MAX_FOG;
		STATE_engine_set_fog(OFF,  220, 220, 180, FOG_MODE_AUTO, fog_visibility, FOG_START_DISTANCE_DEFAULT, 0);
		STATE_font_write_text(ariel18, "Fog: OFF", 0, 50, 255, 0, 0, 100, FONT_WRITE_DEFAULT, INT_MAX);
		return;
	}
	
	
	STATE_engine_set_fog(ON,   220, 220, 180, FOG_MODE_AUTO, fog_visibility, FOG_START_DISTANCE_DEFAULT, 0);

	char fog_text[1000];
	sprintf(fog_text, "Fog intensity: %ld", (int)fog_visibility);
	STATE_font_write_text(ariel18, fog_text, 0, 50, 255, 0, 0, 100, FONT_WRITE_DEFAULT, INT_MAX);

}